Require Import Basics.
Require Import Syntax.

(* Value and environments *)
Inductive UPointer : Set := UPtr : nat -> UPointer.
Inductive LPointer : Set := LPtr : nat -> LPointer.
Inductive MPointer : Set := MPtr : nat -> MPointer.

Axiom Dec_UPointer : forall (x x' : UPointer),  x = x' \/ x <> x'.
Axiom Dec_LPointer : forall (x x' : LPointer),  x = x' \/ x <> x'.
Axiom Dec_MPointer : forall (x x' : MPointer),  x = x' \/ x <> x'.

(* Labeled *)
Inductive LValue : Type :=
  | LVNum  : Label -> Number -> LValue
  | LVFun  : Label -> LFun -> LValue
  | LVRef  : Label -> LRefRead -> LRefWrite -> LValue
  | LVNil  : Label -> LValue
  | LVCons : Label -> LHead -> LTail -> LValue

(* Semantics for LFun will correspond to our functions *)
with LFun : Type := 
  | LLFun : list LPointer -> String -> Expr -> LFun
  | LUFun : list MPointer -> UFun -> UModel -> LModel -> list EffC -> LFun

(* Semantics for LRef will correspond to our functions *)
with LRefRead : Type :=
  | LLRefRead : LPointer -> LRefRead
  | LURefRead : list MPointer -> URefRead -> LModel -> LRefRead

with LRefWrite : Type := 
  | LLRefWrite : LPointer -> LRefWrite
  | LURefWrite : list MPointer -> URefWrite -> LModel (* for getting label to do pc check *) -> UModel -> LRefWrite

with LHead : Type :=
  | LLHead : LValue -> LHead
  | LUHead : list MPointer -> UHead -> LModel -> LHead

with LTail : Type :=
  | LLTail : LValue -> LTail
  | LUTail : list MPointer -> UTail -> LModel -> LTail
                                             
with LItem : Type :=
  | LIValue : LValue -> LItem
  | LIFrame : (String -> option LValue) -> LItem

(* Unlabeled *)
                         
with UValue : Type :=
  | UVNum  : Number -> UValue
  | UVFun  : UFun -> UValue
  | UVRef  : URefRead -> URefWrite -> UValue 
  | UVNil  : UValue
  | UVCons : UHead -> UTail -> UValue

(* Semantics for LFun will correspond to our functions *)
with UFun : Type := 
  | ULFun : Label -> LFun -> Pi -> UFun
  | UUFun : list UPointer -> String -> Expr -> UFun

(* Semantics for LRef will correspond to our functions *)
with URefRead : Type :=
  | ULRefRead : Label -> LRefRead -> Pi -> URefRead
  | UURefRead : UPointer -> URefRead

with URefWrite : Type := 
  | ULRefWrite : Label -> LRefWrite -> Pi -> URefWrite
  | UURefWrite : UPointer -> URefWrite

with UHead : Type :=
  | ULHead : Label -> list MPointer -> LHead -> UModel -> UHead
  | UUHead : UValue -> UHead

with UTail : Type :=
  | ULTail : Label -> list MPointer -> LTail -> UModel -> UTail
  | UUTail : UValue -> UTail

with UItem : Type :=
  | UIValue : UValue -> UItem
  | UIFrame : (String -> option UValue) -> UItem.

(* Model *)

Inductive  MValue : Type :=
  (* label values *)
  | MVLabel     : Label -> MValue
  (* top level models *)
  | MVModel    : LModel -> MValue
  (* effect constraints *)
  | MVRefRead  : UModel -> MValue
  | MVRefWrite : LTerm -> LModel -> MValue
  | MVApp      : LTerm -> LModel -> UModel -> MValue

with MItem : Type := 
  | MIFrame : (String -> option MValue) -> MItem

with Heap : Type :=
  | MkHeap : (LPointer -> LItem) -> (UPointer -> UItem) -> (MPointer -> MItem) -> Heap

with Stack : Type :=
  | MkStack : list LPointer -> list UPointer -> list MPointer -> Stack.

Definition LStack := list LPointer.
Definition UStack := list UPointer.
Definition MStack := list MPointer.
Definition LHeap  := LPointer -> LItem.
Definition UHeap  := UPointer -> UItem.
Definition MHeap  := MPointer -> MItem.
Definition LFrame := String -> option LValue.
Definition UFrame := String -> option UValue.
Definition MFrame := String -> option MValue.

Definition EmptyMFrame (x : String) : option MValue := None.

(* ******************************************************************************** *)

Definition lstack Γ :=
  match Γ with 
  | MkStack lρ uρ mρ => lρ
  end.

Definition ustack Γ :=
  match Γ with 
  | MkStack lρ uρ mρ => uρ
  end.

Definition mstack Γ :=
  match Γ with 
  | MkStack lρ uρ mρ => mρ
  end.

Definition lheap Σ :=
  match Σ with
  | MkHeap lσ uσ mσ => lσ
  end. 

Definition uheap Σ :=
  match Σ with
  | MkHeap lσ uσ mσ => uσ
  end. 

Definition mheap Σ :=
  match Σ with
  | MkHeap lσ uσ mσ => mσ
  end. 
  
Definition LabelOf (v : LValue) : Label := 
  match v with
  | LVNum l _   => l
  | LVFun l _   => l
  | LVRef l _ _ => l
  | LVNil l     => l
  | LVCons l _ _ => l
  end.

Definition Upgrade (l' : Label) (v : LValue) : LValue :=
  match v with
  | LVNum l n   => LVNum (lub l l') n
  | LVFun l F   => LVFun (lub l l') F
  | LVRef l R W => LVRef (lub l l') R W
  | LVNil l     => LVNil (lub l l')
  | LVCons l Hd Tl => LVCons (lub l l') Hd Tl
  end.

(* ******************************************************************************** *)

(* LookupL *************** *) 
Inductive LookupL : LStack -> LHeap -> String -> option LValue -> Prop :=
  | lookupL_ : 
      forall lσ x, 
        LookupL nil lσ x None

  | lookupL1 : 
      forall lσ x p ps ω v, 
        lσ p = LIFrame ω ->
        ω x = Some v ->
        LookupL (cons p ps) lσ x (Some v)

  | lookupL2 : 
      forall lσ x p ps ω v, 
        lσ p = LIFrame ω ->
        ω x = None ->
        LookupL ps lσ x v ->
        LookupL (cons p ps) lσ x v.

(* LookupU *************** *) 
Inductive LookupU : UStack -> UHeap -> String -> option UValue -> Prop :=
  | lookupU_ : 
      forall uσ x, 
        LookupU nil uσ x None

  | lookupU1 : 
      forall uσ x p ps ω v, 
        uσ p = UIFrame ω ->
        ω x = Some v ->
        LookupU (cons p ps) uσ x (Some v)

  | lookupU2 : 
      forall uσ x p ps ω v, 
        uσ p = UIFrame ω ->
        ω x = None ->
        LookupU ps uσ x v ->
        LookupU (cons p ps) uσ x v.


(* LookupM *************** *) 
Inductive LookupM : list MPointer -> MHeap -> String -> option MValue -> Prop :=
  | lookupM_ : 
      forall mσ x, 
        LookupM nil mσ x None

  | lookupM1 : 
      forall mσ x p ps ω v, 
        mσ p = MIFrame ω ->
        ω x = Some v ->
        LookupM (cons p ps) mσ x (Some v)

  | lookupM2 : 
       forall mσ x p ps ω v, 
         mσ p = MIFrame ω ->
         ω x = None ->
         LookupM ps mσ x v ->
         LookupM (cons p ps) mσ x v.

(* FindM *************** *) 
Inductive FindM : list MPointer -> MHeap -> String -> option MPointer -> Prop :=
  | findM_ : 
      forall mσ x, 
        FindM nil mσ x None

  | findM1 : 
      forall mσ x θ θs ω v, 
        mσ θ = MIFrame ω ->
        ω x = Some v ->
        FindM (cons θ θs) mσ x (Some θ)

  | findM2 : 
       forall mσ x p ps ω v, 
         mσ p = MIFrame ω ->
         ω x = None ->
         FindM ps mσ x v ->
         FindM (cons p ps) mσ x v.

(* UpdateM *************** *) 
Inductive UpdateM : Context -> list MPointer -> MHeap -> String -> Label -> MHeap -> Prop :=
  (* Adding under secret control would change from L (not present) to H *)
  | umNotEx : 
      forall p ps σ1 σ2 α l ω1 ω2,
        FindM (cons p ps) σ1 α None -> 
        σ1 p = MIFrame ω1 ->
        ω2 = Update ω1 α (Some (MVLabel l)) ->
        σ2 = Update σ1 p (MIFrame ω2) ->
      UpdateM L (cons p ps) σ1 α l σ2

  | umEx : 
     forall ctx ρ p σ1 σ2 α l l' ω1 ω2,
        FindM ρ σ1 α (Some p) -> 
        σ1 p = MIFrame ω1 ->
        ω1 α = Some (MVLabel l') ->
        CanFlow ctx l' ->
        ω2 = Update ω1 α (Some (MVLabel (lub l l'))) ->
        σ2 = Update σ1 p (MIFrame ω2) ->
      UpdateM ctx ρ σ1 α l σ2.

(* DefineM *************** *) 
(* Only defines in top-most - why we can ignore the context *)
Inductive DefineM : list MPointer -> MHeap -> String -> MValue -> MHeap -> Prop :=
  | dmInsert : 
      forall p ps σ1 σ2 x m ω1 ω2,
        σ1 p = MIFrame ω1 ->
        ω1 x = None ->
        ω2 = Update ω1 x (Some m) ->
        σ2 = Update σ1 p (MIFrame ω2) ->
      DefineM (cons p ps) σ1 x m σ2.

(* Interp_LTerm *************** *) 
Inductive Interp_LTerm : list MPointer -> MHeap -> LTerm -> Label -> Prop :=
  | ilLLabel : 
     forall ρ σ l, 
      Interp_LTerm ρ σ (LLabel l) l

  | ilLVar1 : 
      forall ρ σ α, 
        LookupM ρ σ α None ->
      Interp_LTerm ρ σ (LVar α) L

  | ilLVar2 : 
      forall ρ σ α l, 
        LookupM ρ σ α (Some (MVLabel l)) ->
      Interp_LTerm ρ σ (LVar α) l

 | ilLLub : 
      forall ρ σ κ1 κ2 l1 l2, 
        Interp_LTerm ρ σ κ1 l1 ->
        Interp_LTerm ρ σ κ2 l2 ->
      Interp_LTerm ρ σ (LLub κ1 κ2) (lub l1 l2).

(* Interp_EffC *************** *) 
Inductive Interp_EffC : Context -> Stack -> Heap -> EffC -> Heap -> Prop :=
  | ieECRefRead : 
      forall ctx lρ uρ mρ lσ uσ mσ mσ' π ϕ,
        DefineM mρ mσ π (MVRefRead ϕ) mσ' ->
      Interp_EffC ctx (MkStack lρ uρ mρ) (MkHeap lσ uσ mσ) (ECRefRead π ϕ) (MkHeap lσ uσ mσ') 

  | ieECRefWrite : 
      forall ctx lρ uρ mρ lσ uσ mσ mσ' κ π γ,
        DefineM mρ mσ π (MVRefWrite κ γ) mσ' ->
      Interp_EffC ctx (MkStack lρ uρ mρ) (MkHeap lσ uσ mσ) (ECRefWrite π κ γ) (MkHeap lσ uσ mσ')

  | ieECApp : 
      forall ctx lρ uρ mρ lσ uσ mσ mσ' κ π γ ϕ,
        DefineM mρ mσ π (MVApp κ γ ϕ) mσ' ->
      Interp_EffC ctx (MkStack lρ uρ mρ) (MkHeap lσ uσ mσ) (ECApp π κ γ ϕ) (MkHeap lσ uσ mσ')

  | ieECAsn :
      forall lρ uρ mρ lσ uσ mσ mσ' ctx α cκ κ cl l,
        Interp_LTerm mρ mσ cκ cl ->
        Interp_LTerm mρ mσ κ l ->
        UpdateM (lub ctx cl) mρ mσ α l mσ' ->
      Interp_EffC ctx (MkStack lρ uρ mρ) (MkHeap lσ uσ mσ) (ECAsn cκ α κ) (MkHeap lσ uσ mσ').
 
(* Interp_EffC *************** *) 
Inductive Interp_EffCs : Context -> Stack -> Heap -> list EffC -> Heap -> Prop :=
  | ieNil : 
      forall ctx Γ Σ,
      Interp_EffCs ctx Γ Σ nil Σ

  | ieCons : 
      forall ctx Γ Σ1 Σ2 Σ3 c cs,
        Interp_EffC ctx Γ Σ1 c Σ2 ->
        Interp_EffCs ctx Γ Σ2 cs Σ3 -> 
      Interp_EffCs ctx Γ Σ1 (cons c cs) Σ3.


(* MkLRef, MkURef *************** *) 
Inductive MkLRef : Heap -> LValue -> Heap * LRefRead * LRefWrite -> Prop :=
  | mkLRef :
      forall σl σu σm lv p,
        Fresh σl p ->
      MkLRef (MkHeap σl σu σm) lv (MkHeap (Update σl p (LIValue lv)) σu σm, LLRefRead p, LLRefWrite p).

Inductive MkURef : Heap -> UValue -> Heap * URefRead * URefWrite -> Prop :=
  | mkURef :
      forall σl σu σm uv p,
        Fresh σu p ->
      MkURef (MkHeap σl σu σm) uv (MkHeap σl (Update σu p (UIValue uv)) σm, UURefRead p, UURefWrite p).

(* MkLHead, MkLTail *************** *) 
Definition MkLHead := LLHead.
Definition MkLTail := LLTail.

(* MkUHead, MkUTail *************** *) 
Definition MkUHead := UUHead.
Definition MkUTail := UUTail.

(* MkLFun, MkUFfun *************** *) 
Definition MkLFun := LLFun.
Definition MkUFun := UUFun.


(* Unlabel_LValue  *************** *) 
Inductive Unlabel_LValue : Context -> Stack -> Heap -> LValue -> UModel -> UValue -> Heap -> Prop :=
  | ulLPrim :  
      forall ctx Γ Σ1 Σ2 mσ l n α,
        UpdateM ctx (mstack Γ) (mheap Σ1) α l mσ -> 
        Σ2 = MkHeap (lheap Σ1) (uheap Σ1) mσ -> 
      Unlabel_LValue ctx Γ Σ1 (LVNum l n) (UMPrim α) (UVNum n) Σ2
 
  | ulLFun : 
      forall ctx Γ Σ1 Σ2 mσ l LF α π,
        UpdateM ctx (mstack Γ) (mheap Σ1) α l mσ -> 
        Σ2 = MkHeap (lheap Σ1) (uheap Σ1) mσ -> 
      Unlabel_LValue ctx Γ Σ1 (LVFun l LF) (UMFun α π) (UVFun (ULFun l LF π)) Σ2

  | ulLRef : 
      forall ctx Γ Σ1 Σ2 mσ l LR LW α π,
        UpdateM ctx (mstack Γ) (mheap Σ1) α l mσ -> 
        Σ2 = MkHeap (lheap Σ1) (uheap Σ1) mσ -> 
      Unlabel_LValue ctx Γ Σ1 (LVRef l LR LW) (UMRef α π) (UVRef (ULRefRead l LR π) (ULRefWrite l LW π)) Σ2
 
 | ulLNil : 
      forall ctx Γ Σ1 Σ2 mσ l α ϕ,
        UpdateM ctx (mstack Γ) (mheap Σ1) α l mσ -> 
        Σ2 = MkHeap (lheap Σ1) (uheap Σ1) mσ -> 
      Unlabel_LValue ctx Γ Σ1 (LVNil l) (UMList α ϕ) UVNil Σ2

  | ulLCons : 
      forall ctx Γ Σ1 Σ2 mσ l Hd Tl α ϕ,
        UpdateM ctx (mstack Γ) (mheap Σ1) α l mσ -> 
        Σ2 = MkHeap (lheap Σ1) (uheap Σ1) mσ -> 
      Unlabel_LValue ctx Γ Σ1 (LVCons l Hd Tl) (UMList α ϕ) (UVCons (ULHead l (mstack Γ) Hd ϕ) (ULTail l (mstack Γ) Tl (UMList α ϕ)))  Σ2.

(* Relabel_UValue  *************** *) 
Inductive Relabel_UValue : Stack -> Heap -> UValue -> LModel -> LValue -> Prop :=
  | rlUPrim :
      forall Γ Σ κ l n,
        Interp_LTerm (mstack Γ) (mheap Σ) κ l -> 
      Relabel_UValue Γ Σ (UVNum n) (LMPrim  κ) (LVNum l n)
  
  | rlUFun :
      forall Γ Σ κ ϕ γ ξ l UF,
        Interp_LTerm (mstack Γ) (mheap Σ) κ l -> 
      Relabel_UValue Γ Σ (UVFun UF) (LMFun κ ϕ γ ξ) (LVFun l (LUFun (mstack Γ) UF ϕ γ ξ))

  | rlURef : 
      forall Γ Σ κ ϕ γ l UR UW,
        Interp_LTerm (mstack Γ) (mheap Σ) κ l -> 
      Relabel_UValue Γ Σ (UVRef UR UW) (LMRef κ γ ϕ) (LVRef l (LURefRead (mstack Γ) UR γ) (LURefWrite (mstack Γ) UW γ ϕ))

 | rlUNil :
      forall Γ Σ κ l γ,
        Interp_LTerm (mstack Γ) (mheap Σ) κ l -> 
      Relabel_UValue Γ Σ UVNil (LMList κ γ) (LVNil l)

  | rlUCons : 
      forall Γ Σ κ γ l Hd Tl,
        Interp_LTerm (mstack Γ) (mheap Σ) κ l -> 
      Relabel_UValue Γ Σ (UVCons Hd Tl) (LMList κ γ) (LVCons l (LUHead (mstack Γ) Hd γ) (LUTail (mstack Γ) Tl γ)).


(* LRefReadEval - correspondning to function in paper  *************** *) 
Inductive LRefReadEval : LRefRead -> Context -> Stack -> Heap -> (* *) Heap -> LValue -> Prop :=
  (* called lread in the paper *)
  | levLLRefRead : 
      forall ctx Γ Σ p lv,
        (lheap Σ) p = (LIValue lv) ->
      LRefReadEval (LLRefRead p) ctx Γ Σ (* *) Σ lv

  (* called l-uread in the paper *)
  | levLURefRead :
      forall UR γ ctx mρ Γ Σ1 Σ2 uv lv,
        (URefReadEval UR) ctx Γ Σ1 (* *) Σ2 uv ->
        Relabel_UValue (MkStack (lstack Γ) (ustack Γ) mρ) Σ2 uv γ lv ->
    LRefReadEval (LURefRead mρ UR γ) ctx Γ Σ1 (* *) Σ2 lv

(* URefReadEval - correspondning to function in paper  *************** *) 
with URefReadEval : URefRead -> Context -> Stack -> Heap -> Heap -> UValue -> Prop :=
  (* called uread in the paper *)
  | uevUURefRead : 
      forall ctx Γ Σ p uv,
        (uheap Σ) p = (UIValue uv) ->
      URefReadEval (UURefRead p) ctx Γ Σ (* *) Σ uv 
 
  (* called u-lread in the paper  *)
  | uevULRefRead :
      forall l LR ctx Γ Σ1 π ϕ Σ2 lv Σ3 uv, 
        LookupM (mstack Γ) (mheap Σ1) π (Some (MVRefRead ϕ)) ->
        (LRefReadEval LR) (lub ctx l) Γ Σ1 (* *) Σ2 lv ->
        Unlabel_LValue (lub ctx l) Γ Σ2 lv ϕ uv Σ3 ->
      URefReadEval (ULRefRead l LR π) ctx Γ Σ1 (* *) Σ3 uv.

(* LRefWriteEval - correspondning to function in paper  *************** *) 
Inductive LRefWriteEval : LRefWrite -> Context -> Stack -> Heap -> LValue -> Heap -> Prop :=
  (* called lwrite in the paper *)
  | levLLRefWrite :
      forall ctx Γ Σ1 Σ2 p oldlv lv,
        (lheap Σ1) p = LIValue oldlv ->
        CanFlow ctx (LabelOf oldlv) -> 
        Σ2 = MkHeap (Update (lheap Σ1) p (LIValue (Upgrade ctx lv))) (uheap Σ1) (mheap Σ1) ->
      LRefWriteEval (LLRefWrite p) ctx Γ Σ1 lv Σ2
  
  (* called l-uwrite in the paper *)
  | levLURefWrite :
      forall mρ UW γ ϕ ctx Γ Σ1 oldl lv uv Σ2 Σ3, 
        Interp_LTerm mρ (mheap Σ1) (LTermOf γ) oldl ->
        CanFlow ctx oldl ->
        Unlabel_LValue ctx (MkStack (lstack Γ) (ustack Γ) mρ) Σ1 (Upgrade ctx lv) ϕ uv Σ2 ->
        URefWriteEval UW ctx Γ Σ2 uv Σ3 ->
      LRefWriteEval (LURefWrite mρ UW γ ϕ) ctx  Γ Σ1 lv Σ3

(* URefWriteEval - correspondning to function in paper  *************** *) 
with URefWriteEval : URefWrite -> Context -> Stack -> Heap -> UValue -> Heap -> Prop :=
  (* called uwrite in the paper *) 
  | uevUURefWrite :
       forall ctx Γ Σ1 Σ2 p uv,
        Σ2 = MkHeap (lheap Σ1) (Update (uheap Σ1) p (UIValue uv)) (mheap Σ1) ->
      URefWriteEval (UURefWrite p) ctx Γ Σ1 uv Σ2

  (* called u-lwrite in the paper *)
  | uevULRefWrite : 
      forall ctx l1 l2 LW π κ γ Γ Σ1 Σ2 uv lv,
        LookupM (mstack Γ) (mheap Σ1) π (Some (MVRefWrite κ γ)) ->
        Relabel_UValue Γ Σ1 uv γ lv ->
        Interp_LTerm (mstack Γ) (mheap Σ1) κ l2 ->
        (LRefWriteEval LW) (lub ctx (lub l1 l2)) Γ Σ1 lv Σ2 ->
      URefWriteEval (ULRefWrite l1 LW π) ctx Γ Σ1 uv Σ2.

Inductive LHeadEval : LHead -> Context -> Stack -> Heap -> (* *) Heap -> LValue -> Prop :=
  | levLLHead :
      forall ctx Γ Σ lv,
      LHeadEval (LLHead lv) ctx Γ Σ (* *) Σ lv
  
  | levLUHead :
      forall ctx Γ Σ1 Σ2 UHd mρ uv γ lv,
        UHeadEval UHd ctx Γ Σ1 (* *) Σ2 uv ->
        Relabel_UValue (MkStack (lstack Γ) (ustack Γ) mρ) Σ2 uv γ lv ->
      LHeadEval (LUHead mρ UHd γ) ctx Γ Σ1 Σ2 lv
      
with UHeadEval : UHead -> Context -> Stack -> Heap -> (* *) Heap -> UValue -> Prop :=
  (* called u-lhead in the paper *)
  | uevULHead : 
      forall ctx Γ Σ1 Σ2 Σ3 l LHd mρ lv ϕ uv,
       LHeadEval LHd (lub ctx l) Γ Σ1 (* *) Σ2 lv ->
       Unlabel_LValue (lub ctx l) (MkStack (lstack Γ) (ustack Γ) mρ) Σ2 (Upgrade l lv) ϕ uv Σ3 -> 
     UHeadEval (ULHead l mρ LHd ϕ) ctx Γ Σ1 (* *) Σ3 uv

  | uevUUHead : 
      forall ctx Γ Σ uv,
      UHeadEval (UUHead uv) ctx Γ Σ (* *) Σ uv.

Inductive LTailEval : LTail -> Context -> Stack -> Heap -> (* *) Heap -> LValue -> Prop :=
  | levLLTail : 
      forall ctx Γ Σ lv,
      LTailEval (LLTail lv) ctx Γ Σ (* *) Σ lv
  
  | levLUTail :
      forall ctx Γ Σ1 Σ2 UTl mρ uv γ lv,
        UTailEval UTl ctx Γ Σ1 (* *) Σ2 uv ->
        Relabel_UValue (MkStack (lstack Γ) (ustack Γ) mρ) Σ2 uv γ lv ->
      LTailEval (LUTail mρ UTl γ) ctx Γ Σ1 (* *) Σ2 lv
      
with UTailEval : UTail -> Context -> Stack -> Heap -> (* *) Heap -> UValue -> Prop :=
  | uevULTail : 
      forall ctx Γ Σ1 Σ2 Σ3 l LTl mρ lv ϕ uv,
       LTailEval LTl (lub ctx l) Γ Σ1 (* *) Σ2 lv ->
       Unlabel_LValue (lub ctx l) (MkStack (lstack Γ) (ustack Γ) mρ) Σ2 (Upgrade l lv) ϕ uv Σ3 -> 
     UTailEval (ULTail l mρ LTl ϕ) ctx Γ Σ1 (* *) Σ3 uv

  | uevUUTail : 
      forall ctx Γ Σ uv,
      UTailEval (UUTail uv) ctx Γ Σ (* *) Σ uv.

(* LFunEval - correspondning to function in paper  *************** *) 
Inductive LFunEval : LFun -> Context -> Stack -> Heap -> LValue -> (* *) Heap -> LValue -> Prop :=
  (* called lclos in the paper *)
  | levLLFun :
      forall ctx Γ Σ1 Σ2 Σ3 p σl ρl x e vp vr,
        Fresh (lheap Σ1) p ->
        σl = Update (lheap Σ1) p (LIFrame (Singleton x (Some (Upgrade ctx vp)))) ->
        Σ2 = MkHeap σl (uheap Σ1) (mheap Σ1) -> 
        LEval ctx (MkStack (cons p ρl) (ustack Γ) (mstack Γ)) Σ2 e (* => *) Σ3 vr ->
      LFunEval (LLFun ρl x e) ctx Γ Σ1 vp (* *) Σ3 vr

  (* called l-uclos in the paper *)
  | levLUFun :
      forall mρ UF ϕ γ ξ ctx p Γ1 Γ2 Σ1 Σ2 Σ3 Σ4 Σ5 lv1 lv2 uv1 uv2,
        Fresh (mheap Σ1) p ->
        Γ2 = MkStack (lstack Γ1) (ustack Γ1) (cons p mρ) ->
        Σ2 = MkHeap (lheap Σ1) (uheap Σ1) (Update (mheap Σ1) p (MIFrame EmptyMFrame)) ->
        Unlabel_LValue ctx Γ2 Σ2 lv1 ϕ uv1 Σ3 ->
        Interp_EffCs ctx  Γ2 Σ3 ξ Σ4 ->
        (UFunEval UF) ctx Γ1 Σ4 uv1 (* *) Σ5 uv2 ->
        Relabel_UValue Γ2 Σ5 uv2 γ lv2 ->
      LFunEval (LUFun mρ UF ϕ γ ξ) ctx Γ1 Σ1 lv1 (* *) Σ5 lv2

(* UFunEval - correspondning to function in paper  *************** *) 
with UFunEval : UFun -> Context -> Stack -> Heap -> UValue -> (* *) Heap -> UValue -> Prop :=
  | uevUUFun : 
      forall ctx Γ Σ1 Σ2 Σ3 p σu ρu x e vp vr,
        Fresh (uheap Σ1) p ->
        σu = Update (uheap Σ1) p (UIFrame (Singleton x (Some vp))) ->
        Σ2 = MkHeap (lheap Σ1) σu (mheap Σ1) -> 
        UEval ctx (MkStack (lstack Γ) (cons p ρu) (mstack Γ)) Σ2 e (* => *) Σ3 vr ->
      UFunEval (UUFun ρu x e) ctx Γ Σ1 vp (* *) Σ3 vr

  (* called u-lclos in the paper *)
  | uevULFun : 
      forall l1 l2 LF π ctx Γ Σ1 Σ2 Σ3 κ γ ϕ uv1 uv2 lv1 lv2,
        LookupM (mstack Γ) (mheap Σ1) π (Some (MVApp κ γ ϕ)) ->
        Relabel_UValue Γ Σ1 uv1 γ lv1 ->
        Interp_LTerm (mstack Γ) (mheap Σ1) κ l2 ->
        (LFunEval LF) (lub ctx (lub l1 l2)) Γ Σ1 lv1 (* *) Σ2 lv2 ->
        Unlabel_LValue (lub ctx (lub l1 l2)) Γ Σ2 (Upgrade (lub l1 l2) lv2) ϕ uv2 Σ3 ->
      UFunEval (ULFun l1 LF π) ctx Γ Σ1 uv1 (* *) Σ3 uv2

(* LEval *************** *) 
with LEval : Context -> Stack -> Heap -> Expr -> (* => *) Heap -> LValue -> Prop :=
  | levNum : 
      forall ctx Γ Σ n, 
        LEval ctx Γ Σ (ENum n) (* => *) Σ (LVNum L n)

  | levVar : 
      forall ctx Γ Σ x v, 
        LookupL (lstack Γ) (lheap Σ) x (Some v) -> 
        LEval ctx Γ Σ (EVar x) (* => *) Σ v

  | levIf1 : 
      forall ctx Γ Σ1 Σ2 Σ3 e1 e2 e3 l n v2,
        LEval ctx Γ Σ1 e1 (* => *) Σ2 (LVNum l n) ->
        n <> Num 0 ->
        LEval (lub ctx l) Γ Σ2 e2 (* => *) Σ3 v2 ->
        LEval ctx Γ Σ1 (EIf e1 e2 e3) (* => *) Σ3 (Upgrade l v2)

  | levIf2 : 
      forall ctx Γ Σ1 Σ2 Σ3 e1 e2 e3 l n v3,
        LEval ctx Γ Σ1 e1 (* => *) Σ2 (LVNum l n) ->
        n = Num 0 ->
        LEval (lub ctx l) Γ Σ2 e3 (* => *) Σ3 v3 ->
        LEval ctx Γ Σ1 (EIf e1 e2 e3) (* => *) Σ3 (Upgrade l v3)
 
  | levESeq : 
      forall ctx Γ Σ1 Σ2 Σ3 e1 e2 v1 v2,
        LEval ctx Γ Σ1 e1 (* => *) Σ2 v1 ->
        LEval ctx Γ Σ2 e2 (* => *) Σ3 v2 ->
        LEval ctx Γ Σ1 (ESeq e1 e2) (* => *) Σ3 v2

  | levRef : 
       forall ctx Γ Σ1 Σ2 Σ3 e v LR LW,
         LEval ctx Γ Σ1 e (* => *) Σ2 v ->
         MkLRef Σ2 v (Σ3, LR, LW) ->
         LEval ctx Γ Σ1 (ERef e) (* => *) Σ3 (LVRef L LR LW)

  | levDeref : 
       forall ctx Γ Σ1 Σ2 Σ3 e l LR LW v,
         LEval ctx Γ Σ1 e (* => *) Σ2 (LVRef l LR LW) ->
         (LRefReadEval LR) (lub ctx l) Γ Σ2 (* *) Σ3 v ->
         LEval ctx Γ Σ1 (EDeref e) (* => *) Σ3 (Upgrade l v)

  | levAsn : 
      forall ctx Γ Σ1 Σ2 Σ3 Σ4 e1 e2 l R W v,
        LEval ctx Γ Σ1 e1 (* => *) Σ2 (LVRef l R W) ->
        LEval ctx Γ Σ2 e2 (* => *) Σ3 v ->
        (LRefWriteEval W) (lub ctx l) Γ Σ3 v Σ4 ->
        LEval ctx Γ Σ1 (EAsn e1 e2) (* => *) Σ4 v

  | levFun : 
      forall ctx Γ Σ x e F,
        F = MkLFun (lstack Γ) x e -> 
        LEval ctx Γ Σ (EFun x e) (* => *) Σ (LVFun L F)

  | levApp : 
      forall ctx Γ Σ1 Σ2 Σ3 Σ4 ef e l F lv1 lv2,
        LEval ctx Γ Σ1 ef (* => *) Σ2 (LVFun l F) ->
        LEval ctx Γ Σ2 e (* => *) Σ3 lv1 ->
        (LFunEval F) (lub ctx l) Γ Σ3 lv1 (* *) Σ4 lv2 ->
        LEval ctx Γ Σ1 (EApp ef e) (* => *) Σ4 (Upgrade l lv2)
 
  | levLib : 
      forall ctx Γ Σ f uv γ lv,
        LookupU (ustack Γ) (uheap Σ) f (Some uv) ->
        LookupM (mstack Γ) (mheap Σ) f (Some (MVModel γ)) ->
        Relabel_UValue Γ Σ uv γ lv ->
        LEval ctx Γ Σ (ELib f) (* => *) Σ lv

  | levNil :
      forall ctx Γ Σ, 
      LEval ctx Γ Σ ENil (* => *) Σ (LVNil L)

  | levCons : 
      forall ctx Γ Σ1 Σ2 Σ3 e1 e2 v1 v2,
        LEval ctx Γ Σ1 e1 (* => *) Σ2 v1 ->
        LEval ctx Γ Σ2 e2 (* => *) Σ3 v2 ->
        LEval ctx Γ Σ1 (ECons e1 e2) (* => *) Σ3 (LVCons L (MkLHead v1) (MkLTail v2))

  | levHead :
      forall ctx Γ Σ1 Σ2 Σ3 LHd LTl l e lv, 
        LEval ctx Γ Σ1 e (* => *) Σ2 (LVCons l LHd LTl) ->
        LHeadEval LHd (lub ctx l) Γ Σ2 (* *) Σ3 lv ->
      LEval ctx Γ Σ1 (EHead e) (* => *) Σ3 (Upgrade l lv)

 | levTail :
      forall ctx Γ Σ1 Σ2 Σ3 LHd LTl l e lv, 
        LEval ctx Γ Σ1 e (* => *) Σ2 (LVCons l LHd LTl) ->
        LTailEval LTl (lub ctx l) Γ Σ2 (* *) Σ3 lv ->
      LEval ctx Γ Σ1 (ETail e) (* => *) Σ3 (Upgrade l lv)

(* UEval *************** *) 
with UEval : Context -> Stack -> Heap -> Expr -> (* => *) Heap -> UValue -> Prop :=
  | uevNum : forall pc Γ Σ n, UEval pc Γ Σ (ENum n) (* => *) Σ (UVNum n)

  | uevVar : 
      forall ctx Γ Σ x v, 
        LookupU (ustack Γ) (uheap Σ) x (Some v) -> 
        UEval ctx Γ Σ (EVar x) (* => *) Σ v

  | uevIf1 : 
      forall ctx Γ Σ1 Σ2 Σ3 e1 e2 e3 n v2,
        UEval ctx Γ Σ1 e1 (* => *) Σ2 (UVNum n) ->
        n <> Num 0 ->
        UEval ctx Γ Σ2 e2 (* => *) Σ3 v2 ->
        UEval ctx Γ Σ1 (EIf e1 e2 e3) (* => *) Σ3 v2

  | uevIf2 : 
      forall ctx Γ Σ1 Σ2 Σ3 e1 e2 e3 n v3,
        UEval ctx Γ Σ1 e1 (* => *) Σ2 (UVNum n) ->
        n = Num 0 ->
        UEval ctx Γ Σ2 e3 (* => *) Σ3 v3 ->
        UEval ctx Γ Σ1 (EIf e1 e2 e3) (* => *) Σ3 v3
 
  | uevESeq : 
      forall ctx Γ Σ1 Σ2 Σ3 e1 e2 v1 v2,
        UEval ctx Γ Σ1 e1 (* => *) Σ2 v1 ->
        UEval ctx Γ Σ2 e2 (* => *) Σ3 v2 ->
        UEval ctx Γ Σ1 (ESeq e1 e2) (* => *) Σ3 v2

  | uevRef : 
       forall ctx Γ Σ1 Σ2 Σ3 e v UR UW,
         UEval ctx Γ Σ1 e (* => *) Σ2 v ->
         MkURef Σ2 v (Σ3, UR, UW) ->
         UEval ctx Γ Σ1 (ERef e) (* => *) Σ3 (UVRef UR UW)

  | uevDeref : 
       forall ctx Γ Σ1 Σ2 Σ3 e UR UW v,
         UEval ctx Γ Σ1 e (* => *) Σ2 (UVRef UR UW) ->
         (URefReadEval UR) ctx Γ Σ2 (* *) Σ3 v ->
         UEval ctx Γ Σ1 (EDeref e) (* => *) Σ3 v

  | uevAsn : 
      forall ctx Γ Σ1 Σ2 Σ3 Σ4 e1 e2 UR UW v,
        UEval ctx Γ Σ1 e1 (* => *) Σ2 (UVRef UR UW) ->
        UEval ctx Γ Σ2 e2 (* => *) Σ3 v ->
        (URefWriteEval UW) ctx Γ Σ3 v Σ4 ->
        UEval ctx Γ Σ1 (EAsn e1 e2) (* => *) Σ4 v

  | uevFun : 
      forall ctx Γ Σ x e F,
        F = MkUFun (ustack Γ) x e -> 
        UEval ctx Γ Σ (EFun x e) (* => *) Σ (UVFun F)

  | uevApp : 
      forall ctx Γ Σ1 Σ2 Σ3 Σ4 ef e F lv1 lv2,
        UEval ctx Γ Σ1 ef (* => *) Σ2 (UVFun F) ->
        UEval ctx Γ Σ2 e (* => *) Σ3 lv1 ->
        (UFunEval F) ctx Γ Σ3 lv1 (* *) Σ4 lv2 ->
        UEval ctx Γ Σ1 (EApp ef e) (* => *) Σ4 lv2

  | uevNil :
      forall ctx Γ Σ, 
      UEval ctx Γ Σ ENil (* => *) Σ UVNil

  | uevCons : 
      forall ctx Γ Σ1 Σ2 Σ3 e1 e2 v1 v2,
        UEval ctx Γ Σ1 e1 (* => *) Σ2 v1 ->
        UEval ctx Γ Σ2 e2 (* => *) Σ3 v2 ->
        UEval ctx Γ Σ1 (ECons e1 e2) (* => *) Σ3 (UVCons (MkUHead v1) (MkUTail v2))

  | uevHead :
      forall ctx Γ Σ1 Σ2 Σ3 UHd UTl e uv, 
        UEval ctx Γ Σ1 e (* => *) Σ2 (UVCons UHd UTl) ->
        UHeadEval UHd ctx Γ Σ2 (* *) Σ3 uv ->
      UEval ctx Γ Σ1 (EHead e) (* => *) Σ3 uv

 | uevTail :
      forall ctx Γ Σ1 Σ2 Σ3 UHd UTl e uv, 
        UEval ctx Γ Σ1 e (* => *) Σ2 (UVCons UHd UTl) ->
        UTailEval UTl ctx Γ Σ2 (* *) Σ3 uv ->
      UEval ctx Γ Σ1 (ETail e) (* => *) Σ3 uv.

Scheme LRefReadEval_mutind := Induction for LRefReadEval Sort Prop 
  with URefReadEval_mutind := Induction for URefReadEval Sort Prop.

Scheme LRefWriteEval_mutind := Induction for LRefWriteEval Sort Prop
  with URefWriteEval_mutind := Induction for URefWriteEval Sort Prop.

Scheme LHeadEval_mutind := Induction for LHeadEval Sort Prop
  with UHeadEval_mutind := Induction for UHeadEval Sort Prop.

Scheme LTailEval_mutind := Induction for LTailEval Sort Prop
  with UTailEval_mutind := Induction for UTailEval Sort Prop.

Scheme LFunEval_mutind := Induction for LFunEval Sort Prop
  with UFunEval_mutind := Induction for UFunEval Sort Prop
  with LEval_mutind := Induction for LEval Sort Prop
  with UEval_mutind := Induction for UEval Sort Prop.

(* Proper mutual induction hypotheses *)

Lemma RefReadEval_mutind : 
  
forall
         (P : forall (l : LRefRead) (c : Context) (s : Stack) (h h0 : Heap) (l0 : LValue), LRefReadEval l c s h h0 l0 -> Prop)
         (P0 : forall (u : URefRead) (c : Context) (s : Stack) (h h0 : Heap) (u0 : UValue), URefReadEval u c s h h0 u0 -> Prop),
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (p : LPointer) (lv : LValue) (e : lheap Σ p = LIValue lv),
        P (LLRefRead p) ctx Γ Σ Σ lv (levLLRefRead ctx Γ Σ p lv e)) ->
       (forall (UR : URefRead) (γ : LModel) (ctx : Context) (mρ : list MPointer) (Γ : Stack) (Σ1 Σ2 : Heap) 
          (uv : UValue) (lv : LValue) (u : URefReadEval UR ctx Γ Σ1 Σ2 uv),
        P0 UR ctx Γ Σ1 Σ2 uv u ->
        forall r : Relabel_UValue (MkStack (lstack Γ) (ustack Γ) mρ) Σ2 uv γ lv,
        P (LURefRead mρ UR γ) ctx Γ Σ1 Σ2 lv (levLURefRead UR γ ctx mρ Γ Σ1 Σ2 uv lv u r)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (p : UPointer) (uv : UValue) (e : uheap Σ p = UIValue uv),
        P0 (UURefRead p) ctx Γ Σ Σ uv (uevUURefRead ctx Γ Σ p uv e)) ->
       (forall (l : Label) (LR : LRefRead) (ctx : Label) (Γ : Stack) (Σ1 : Heap) (π : String) (ϕ : UModel) 
          (Σ2 : Heap) (lv : LValue) (Σ3 : Heap) (uv : UValue) (l0 : LookupM (mstack Γ) (mheap Σ1) π (Some (MVRefRead ϕ)))
          (l1 : LRefReadEval LR (lub ctx l) Γ Σ1 Σ2 lv),
        P LR (lub ctx l) Γ Σ1 Σ2 lv l1 ->
        forall u : Unlabel_LValue (lub ctx l) Γ Σ2 lv ϕ uv Σ3,
        P0 (ULRefRead l LR π) ctx Γ Σ1 Σ3 uv (uevULRefRead l LR ctx Γ Σ1 π ϕ Σ2 lv Σ3 uv l0 l1 u)) ->
       (forall (l : LRefRead) (c : Context) (s : Stack) (h h0 : Heap) (l0 : LValue) (l1 : LRefReadEval l c s h h0 l0),
       P l c s h h0 l0 l1) /\

       (forall (u : URefRead) (c : Context) (s : Stack) (h h0 : Heap) (u0 : UValue) (u1 : URefReadEval u c s h h0 u0),
       P0 u c s h h0 u0 u1).
Proof.
intros.
split.
apply LRefReadEval_mutind with (P:=P)(P0:=P0); assumption.
apply URefReadEval_mutind with (P:=P)(P0:=P0); assumption.
Qed.

Lemma RefWriteEval_mutind :
  forall
         (P : forall (l : LRefWrite) (c : Context) (s : Stack) (h : Heap) (l0 : LValue) (h0 : Heap),
              LRefWriteEval l c s h l0 h0 -> Prop)
         (P0 : forall (u : URefWrite) (c : Context) (s : Stack) (h : Heap) (u0 : UValue) (h0 : Heap),
               URefWriteEval u c s h u0 h0 -> Prop),
       (forall (ctx : Label) (Γ : Stack) (Σ1 Σ2 : Heap) (p : LPointer) (oldlv lv : LValue) (e : lheap Σ1 p = LIValue oldlv)
          (c : CanFlow ctx (LabelOf oldlv))
          (e0 : Σ2 = MkHeap (Update (lheap Σ1) p (LIValue (Upgrade ctx lv))) (uheap Σ1) (mheap Σ1)),
        P (LLRefWrite p) ctx Γ Σ1 lv Σ2 (levLLRefWrite ctx Γ Σ1 Σ2 p oldlv lv e c e0)) ->
       (forall (mρ : list MPointer) (UW : URefWrite) (γ : LModel) (ϕ : UModel) (ctx : Label) (Γ : Stack) 
          (Σ1 : Heap) (oldl : Label) (lv : LValue) (uv : UValue) (Σ2 Σ3 : Heap)
          (i : Interp_LTerm mρ (mheap Σ1) (LTermOf γ) oldl) (c : CanFlow ctx oldl)
          (u : Unlabel_LValue ctx (MkStack (lstack Γ) (ustack Γ) mρ) Σ1 (Upgrade ctx lv) ϕ uv Σ2)
          (u0 : URefWriteEval UW ctx Γ Σ2 uv Σ3),
        P0 UW ctx Γ Σ2 uv Σ3 u0 ->
        P (LURefWrite mρ UW γ ϕ) ctx Γ Σ1 lv Σ3 (levLURefWrite mρ UW γ ϕ ctx Γ Σ1 oldl lv uv Σ2 Σ3 i c u u0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 : Heap) (p : UPointer) (uv : UValue)
          (e : Σ2 = MkHeap (lheap Σ1) (Update (uheap Σ1) p (UIValue uv)) (mheap Σ1)),
        P0 (UURefWrite p) ctx Γ Σ1 uv Σ2 (uevUURefWrite ctx Γ Σ1 Σ2 p uv e)) ->
       (forall (ctx l1 l2 : Label) (LW : LRefWrite) (π : String) (κ : LTerm) (γ : LModel) (Γ : Stack) 
          (Σ1 Σ2 : Heap) (uv : UValue) (lv : LValue) (l : LookupM (mstack Γ) (mheap Σ1) π (Some (MVRefWrite κ γ)))
          (r : Relabel_UValue Γ Σ1 uv γ lv) (i : Interp_LTerm (mstack Γ) (mheap Σ1) κ l2)
          (l0 : LRefWriteEval LW (lub ctx (lub l1 l2)) Γ Σ1 lv Σ2),
        P LW (lub ctx (lub l1 l2)) Γ Σ1 lv Σ2 l0 ->
        P0 (ULRefWrite l1 LW π) ctx Γ Σ1 uv Σ2 (uevULRefWrite ctx l1 l2 LW π κ γ Γ Σ1 Σ2 uv lv l r i l0)) ->
(       forall (l : LRefWrite) (c : Context) (s : Stack) (h : Heap) (l0 : LValue) (h0 : Heap) (l1 : LRefWriteEval l c s h l0 h0),
       P l c s h l0 h0 l1 )/\
(forall (u : URefWrite) (c : Context) (s : Stack) (h : Heap) (u0 : UValue) (h0 : Heap)
         (u1 : URefWriteEval u c s h u0 h0), P0 u c s h u0 h0 u1).
Proof.
intros.
split.
apply LRefWriteEval_mutind with (P:=P)(P0:=P0); assumption.
apply URefWriteEval_mutind with (P:=P)(P0:=P0); assumption.
Qed.


Lemma HeadEval_mutind :
   forall (P : forall (l : LHead) (c : Context) (s : Stack) (h h0 : Heap) (l0 : LValue), LHeadEval l c s h h0 l0 -> Prop)
         (P0 : forall (u : UHead) (c : Context) (s : Stack) (h h0 : Heap) (u0 : UValue), UHeadEval u c s h h0 u0 -> Prop),
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (lv : LValue), P (LLHead lv) ctx Γ Σ Σ lv (levLLHead ctx Γ Σ lv)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 : Heap) (UHd : UHead) (mρ : list MPointer) (uv : UValue) 
          (γ : LModel) (lv : LValue) (u : UHeadEval UHd ctx Γ Σ1 Σ2 uv),
        P0 UHd ctx Γ Σ1 Σ2 uv u ->
        forall r : Relabel_UValue (MkStack (lstack Γ) (ustack Γ) mρ) Σ2 uv γ lv,
        P (LUHead mρ UHd γ) ctx Γ Σ1 Σ2 lv (levLUHead ctx Γ Σ1 Σ2 UHd mρ uv γ lv u r)) ->
       (forall (ctx : Label) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (l : Label) (LHd : LHead) (mρ : list MPointer) 
          (lv : LValue) (ϕ : UModel) (uv : UValue) (l0 : LHeadEval LHd (lub ctx l) Γ Σ1 Σ2 lv),
        P LHd (lub ctx l) Γ Σ1 Σ2 lv l0 ->
        forall u : Unlabel_LValue (lub ctx l) (MkStack (lstack Γ) (ustack Γ) mρ) Σ2 (Upgrade l lv) ϕ uv Σ3,
        P0 (ULHead l mρ LHd ϕ) ctx Γ Σ1 Σ3 uv (uevULHead ctx Γ Σ1 Σ2 Σ3 l LHd mρ lv ϕ uv l0 u)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (uv : UValue), P0 (UUHead uv) ctx Γ Σ Σ uv (uevUUHead ctx Γ Σ uv)) ->
      ( forall (l : LHead) (c : Context) (s : Stack) (h h0 : Heap) (l0 : LValue) (l1 : LHeadEval l c s h h0 l0),
       P l c s h h0 l0 l1) /\

 (forall (u : UHead) (c : Context) (s : Stack) (h h0 : Heap) (u0 : UValue)
         (u1 : UHeadEval u c s h h0 u0), P0 u c s h h0 u0 u1).
Proof.
intros.
split.
apply LHeadEval_mutind with (P:=P)(P0:=P0); assumption.
apply UHeadEval_mutind with (P:=P)(P0:=P0); assumption.
Qed.


Lemma TailEval_mutind :
forall (P : forall (l : LTail) (c : Context) (s : Stack) (h h0 : Heap) (l0 : LValue), LTailEval l c s h h0 l0 -> Prop)
         (P0 : forall (u : UTail) (c : Context) (s : Stack) (h h0 : Heap) (u0 : UValue), UTailEval u c s h h0 u0 -> Prop),
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (lv : LValue), P (LLTail lv) ctx Γ Σ Σ lv (levLLTail ctx Γ Σ lv)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 : Heap) (UTl : UTail) (mρ : list MPointer) (uv : UValue) 
          (γ : LModel) (lv : LValue) (u : UTailEval UTl ctx Γ Σ1 Σ2 uv),
        P0 UTl ctx Γ Σ1 Σ2 uv u ->
        forall r : Relabel_UValue (MkStack (lstack Γ) (ustack Γ) mρ) Σ2 uv γ lv,
        P (LUTail mρ UTl γ) ctx Γ Σ1 Σ2 lv (levLUTail ctx Γ Σ1 Σ2 UTl mρ uv γ lv u r)) ->
       (forall (ctx : Label) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (l : Label) (LTl : LTail) (mρ : list MPointer) 
          (lv : LValue) (ϕ : UModel) (uv : UValue) (l0 : LTailEval LTl (lub ctx l) Γ Σ1 Σ2 lv),
        P LTl (lub ctx l) Γ Σ1 Σ2 lv l0 ->
        forall u : Unlabel_LValue (lub ctx l) (MkStack (lstack Γ) (ustack Γ) mρ) Σ2 (Upgrade l lv) ϕ uv Σ3,
        P0 (ULTail l mρ LTl ϕ) ctx Γ Σ1 Σ3 uv (uevULTail ctx Γ Σ1 Σ2 Σ3 l LTl mρ lv ϕ uv l0 u)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (uv : UValue), P0 (UUTail uv) ctx Γ Σ Σ uv (uevUUTail ctx Γ Σ uv)) ->
 (      forall (l : LTail) (c : Context) (s : Stack) (h h0 : Heap) (l0 : LValue) (l1 : LTailEval l c s h h0 l0),
       P l c s h h0 l0 l1) /\

(forall (u : UTail) (c : Context) (s : Stack) (h h0 : Heap) (u0 : UValue)
         (u1 : UTailEval u c s h h0 u0), P0 u c s h h0 u0 u1).
Proof.
intros.
split.
apply LTailEval_mutind with (P:=P)(P0:=P0); assumption.
apply UTailEval_mutind with (P:=P)(P0:=P0); assumption.
Qed.

Lemma Eval_mutind :  
 forall
         (P : forall (l : LFun) (c : Context) (s : Stack) (h : Heap) (l0 : LValue) (h0 : Heap) (l1 : LValue),
              LFunEval l c s h l0 h0 l1 -> Prop)
         (P0 : forall (u : UFun) (c : Context) (s : Stack) (h : Heap) (u0 : UValue) (h0 : Heap) (u1 : UValue),
               UFunEval u c s h u0 h0 u1 -> Prop)
         (P1 : forall (c : Context) (s : Stack) (h : Heap) (e : Expr) (h0 : Heap) (l : LValue),
               LEval c s h e h0 l -> Prop)
         (P2 : forall (c : Context) (s : Stack) (h : Heap) (e : Expr) (h0 : Heap) (u : UValue),
               UEval c s h e h0 u -> Prop),
       (forall (ctx : Label) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (p : LPointer) (σl : LPointer -> LItem)
          (ρl : list LPointer) (x : String) (e : Expr) (vp vr : LValue) (f : Fresh (lheap Σ1) p)
          (e0 : σl = Update (lheap Σ1) p (LIFrame (Singleton x (Some (Upgrade ctx vp)))))
          (e1 : Σ2 = MkHeap σl (uheap Σ1) (mheap Σ1))
          (l : LEval ctx (MkStack (p :: ρl) (ustack Γ) (mstack Γ)) Σ2 e Σ3 vr),
        P1 ctx (MkStack (p :: ρl) (ustack Γ) (mstack Γ)) Σ2 e Σ3 vr l ->
        P (LLFun ρl x e) ctx Γ Σ1 vp Σ3 vr (levLLFun ctx Γ Σ1 Σ2 Σ3 p σl ρl x e vp vr f e0 e1 l)) ->
       (forall (mρ : list MPointer) (UF : UFun) (ϕ : UModel) (γ : LModel) (ξ : list EffC) 
          (ctx : Context) (p : MPointer) (Γ1 Γ2 : Stack) (Σ1 Σ2 Σ3 Σ4 Σ5 : Heap) (lv1 lv2 : LValue)
          (uv1 uv2 : UValue) (f0 : Fresh (mheap Σ1) p) (e : Γ2 = MkStack (lstack Γ1) (ustack Γ1) (p :: mρ))
          (e0 : Σ2 = MkHeap (lheap Σ1) (uheap Σ1) (Update (mheap Σ1) p (MIFrame EmptyMFrame)))
          (u : Unlabel_LValue ctx Γ2 Σ2 lv1 ϕ uv1 Σ3) (i : Interp_EffCs ctx Γ2 Σ3 ξ Σ4)
          (u0 : UFunEval UF ctx Γ1 Σ4 uv1 Σ5 uv2),
        P0 UF ctx Γ1 Σ4 uv1 Σ5 uv2 u0 ->
        forall r : Relabel_UValue Γ2 Σ5 uv2 γ lv2,
        P (LUFun mρ UF ϕ γ ξ) ctx Γ1 Σ1 lv1 Σ5 lv2
          (levLUFun mρ UF ϕ γ ξ ctx p Γ1 Γ2 Σ1 Σ2 Σ3 Σ4 Σ5 lv1 lv2 uv1 uv2 f0 e e0 u i u0 r)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (p : UPointer) (σu : UPointer -> UItem)
          (ρu : list UPointer) (x : String) (e : Expr) (vp vr : UValue) (f1 : Fresh (uheap Σ1) p)
          (e0 : σu = Update (uheap Σ1) p (UIFrame (Singleton x (Some vp))))
          (e1 : Σ2 = MkHeap (lheap Σ1) σu (mheap Σ1))
          (u : UEval ctx (MkStack (lstack Γ) (p :: ρu) (mstack Γ)) Σ2 e Σ3 vr),
        P2 ctx (MkStack (lstack Γ) (p :: ρu) (mstack Γ)) Σ2 e Σ3 vr u ->
        P0 (UUFun ρu x e) ctx Γ Σ1 vp Σ3 vr (uevUUFun ctx Γ Σ1 Σ2 Σ3 p σu ρu x e vp vr f1 e0 e1 u)) ->
       (forall (l1 l2 : Label) (LF : LFun) (π : String) (ctx : Label) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) 
          (κ : LTerm) (γ : LModel) (ϕ : UModel) (uv1 uv2 : UValue) (lv1 lv2 : LValue)
          (l : LookupM (mstack Γ) (mheap Σ1) π (Some (MVApp κ γ ϕ))) (r : Relabel_UValue Γ Σ1 uv1 γ lv1)
          (i : Interp_LTerm (mstack Γ) (mheap Σ1) κ l2) (l0 : LFunEval LF (lub ctx (lub l1 l2)) Γ Σ1 lv1 Σ2 lv2),
        P LF (lub ctx (lub l1 l2)) Γ Σ1 lv1 Σ2 lv2 l0 ->
        forall u : Unlabel_LValue (lub ctx (lub l1 l2)) Γ Σ2 (Upgrade (lub l1 l2) lv2) ϕ uv2 Σ3,
        P0 (ULFun l1 LF π) ctx Γ Σ1 uv1 Σ3 uv2 (uevULFun l1 l2 LF π ctx Γ Σ1 Σ2 Σ3 κ γ ϕ uv1 uv2 lv1 lv2 l r i l0 u)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (n : Number),
        P1 ctx Γ Σ (ENum n) Σ (LVNum L n) (levNum ctx Γ Σ n)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (x : String) (v : LValue)
          (l : LookupL (lstack Γ) (lheap Σ) x (Some v)), P1 ctx Γ Σ (EVar x) Σ v (levVar ctx Γ Σ x v l)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e1 e2 e3 : Expr) (l : Label) 
          (n : Number) (v2 : LValue) (l0 : LEval ctx Γ Σ1 e1 Σ2 (LVNum l n)),
        P1 ctx Γ Σ1 e1 Σ2 (LVNum l n) l0 ->
        forall (n0 : n <> Num 0) (l1 : LEval (lub ctx l) Γ Σ2 e2 Σ3 v2),
        P1 (lub ctx l) Γ Σ2 e2 Σ3 v2 l1 ->
        P1 ctx Γ Σ1 (EIf e1 e2 e3) Σ3 (Upgrade l v2) (levIf1 ctx Γ Σ1 Σ2 Σ3 e1 e2 e3 l n v2 l0 n0 l1)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e1 e2 e3 : Expr) (l : Label) 
          (n : Number) (v3 : LValue) (l0 : LEval ctx Γ Σ1 e1 Σ2 (LVNum l n)),
        P1 ctx Γ Σ1 e1 Σ2 (LVNum l n) l0 ->
        forall (e : n = Num 0) (l1 : LEval (lub ctx l) Γ Σ2 e3 Σ3 v3),
        P1 (lub ctx l) Γ Σ2 e3 Σ3 v3 l1 ->
        P1 ctx Γ Σ1 (EIf e1 e2 e3) Σ3 (Upgrade l v3) (levIf2 ctx Γ Σ1 Σ2 Σ3 e1 e2 e3 l n v3 l0 e l1)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e1 e2 : Expr) (v1 v2 : LValue)
          (l : LEval ctx Γ Σ1 e1 Σ2 v1),
        P1 ctx Γ Σ1 e1 Σ2 v1 l ->
        forall l0 : LEval ctx Γ Σ2 e2 Σ3 v2,
        P1 ctx Γ Σ2 e2 Σ3 v2 l0 -> P1 ctx Γ Σ1 (ESeq e1 e2) Σ3 v2 (levESeq ctx Γ Σ1 Σ2 Σ3 e1 e2 v1 v2 l l0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e : Expr) (v : LValue) (LR : LRefRead) 
          (LW : LRefWrite) (l : LEval ctx Γ Σ1 e Σ2 v),
        P1 ctx Γ Σ1 e Σ2 v l ->
        forall m : MkLRef Σ2 v (Σ3, LR, LW),
        P1 ctx Γ Σ1 (ERef e) Σ3 (LVRef L LR LW) (levRef ctx Γ Σ1 Σ2 Σ3 e v LR LW l m)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e : Expr) (l : Label) (LR : LRefRead) 
          (LW : LRefWrite) (v : LValue) (l0 : LEval ctx Γ Σ1 e Σ2 (LVRef l LR LW)),
        P1 ctx Γ Σ1 e Σ2 (LVRef l LR LW) l0 ->
        forall l1 : LRefReadEval LR (lub ctx l) Γ Σ2 Σ3 v,
        P1 ctx Γ Σ1 (EDeref e) Σ3 (Upgrade l v) (levDeref ctx Γ Σ1 Σ2 Σ3 e l LR LW v l0 l1)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 Σ4 : Heap) (e1 e2 : Expr) (l : Label) 
          (R : LRefRead) (W : LRefWrite) (v : LValue) (l0 : LEval ctx Γ Σ1 e1 Σ2 (LVRef l R W)),
        P1 ctx Γ Σ1 e1 Σ2 (LVRef l R W) l0 ->
        forall l1 : LEval ctx Γ Σ2 e2 Σ3 v,
        P1 ctx Γ Σ2 e2 Σ3 v l1 ->
        forall l2 : LRefWriteEval W (lub ctx l) Γ Σ3 v Σ4,
        P1 ctx Γ Σ1 (EAsn e1 e2) Σ4 v (levAsn ctx Γ Σ1 Σ2 Σ3 Σ4 e1 e2 l R W v l0 l1 l2)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (x : String) (e : Expr) (F : LFun)
          (e0 : F = MkLFun (lstack Γ) x e), P1 ctx Γ Σ (EFun x e) Σ (LVFun L F) (levFun ctx Γ Σ x e F e0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 Σ4 : Heap) (ef e : Expr) (l : Label) 
          (F : LFun) (lv1 lv2 : LValue) (l0 : LEval ctx Γ Σ1 ef Σ2 (LVFun l F)),
        P1 ctx Γ Σ1 ef Σ2 (LVFun l F) l0 ->
        forall l1 : LEval ctx Γ Σ2 e Σ3 lv1,
        P1 ctx Γ Σ2 e Σ3 lv1 l1 ->
        forall l2 : LFunEval F (lub ctx l) Γ Σ3 lv1 Σ4 lv2,
        P F (lub ctx l) Γ Σ3 lv1 Σ4 lv2 l2 ->
        P1 ctx Γ Σ1 (EApp ef e) Σ4 (Upgrade l lv2) (levApp ctx Γ Σ1 Σ2 Σ3 Σ4 ef e l F lv1 lv2 l0 l1 l2)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (f13 : String) (uv : UValue) (γ : LModel) 
          (lv : LValue) (l : LookupU (ustack Γ) (uheap Σ) f13 (Some uv))
          (l0 : LookupM (mstack Γ) (mheap Σ) f13 (Some (MVModel γ))) (r : Relabel_UValue Γ Σ uv γ lv),
        P1 ctx Γ Σ (ELib f13) Σ lv (levLib ctx Γ Σ f13 uv γ lv l l0 r)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap), P1 ctx Γ Σ ENil Σ (LVNil L) (levNil ctx Γ Σ)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e1 e2 : Expr) (v1 v2 : LValue)
          (l : LEval ctx Γ Σ1 e1 Σ2 v1),
        P1 ctx Γ Σ1 e1 Σ2 v1 l ->
        forall l0 : LEval ctx Γ Σ2 e2 Σ3 v2,
        P1 ctx Γ Σ2 e2 Σ3 v2 l0 ->
        P1 ctx Γ Σ1 (ECons e1 e2) Σ3 (LVCons L (MkLHead v1) (MkLTail v2)) (levCons ctx Γ Σ1 Σ2 Σ3 e1 e2 v1 v2 l l0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (LHd : LHead) (LTl : LTail) 
          (l : Label) (e : Expr) (lv : LValue) (l0 : LEval ctx Γ Σ1 e Σ2 (LVCons l LHd LTl)),
        P1 ctx Γ Σ1 e Σ2 (LVCons l LHd LTl) l0 ->
        forall l1 : LHeadEval LHd (lub ctx l) Γ Σ2 Σ3 lv,
        P1 ctx Γ Σ1 (EHead e) Σ3 (Upgrade l lv) (levHead ctx Γ Σ1 Σ2 Σ3 LHd LTl l e lv l0 l1)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (LHd : LHead) (LTl : LTail) 
          (l : Label) (e : Expr) (lv : LValue) (l0 : LEval ctx Γ Σ1 e Σ2 (LVCons l LHd LTl)),
        P1 ctx Γ Σ1 e Σ2 (LVCons l LHd LTl) l0 ->
        forall l1 : LTailEval LTl (lub ctx l) Γ Σ2 Σ3 lv,
        P1 ctx Γ Σ1 (ETail e) Σ3 (Upgrade l lv) (levTail ctx Γ Σ1 Σ2 Σ3 LHd LTl l e lv l0 l1)) ->
       (forall (pc : Context) (Γ : Stack) (Σ : Heap) (n : Number), P2 pc Γ Σ (ENum n) Σ (UVNum n) (uevNum pc Γ Σ n)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (x : String) (v : UValue)
          (l : LookupU (ustack Γ) (uheap Σ) x (Some v)), P2 ctx Γ Σ (EVar x) Σ v (uevVar ctx Γ Σ x v l)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e1 e2 e3 : Expr) (n : Number) 
          (v2 : UValue) (u : UEval ctx Γ Σ1 e1 Σ2 (UVNum n)),
        P2 ctx Γ Σ1 e1 Σ2 (UVNum n) u ->
        forall (n0 : n <> Num 0) (u0 : UEval ctx Γ Σ2 e2 Σ3 v2),
        P2 ctx Γ Σ2 e2 Σ3 v2 u0 -> P2 ctx Γ Σ1 (EIf e1 e2 e3) Σ3 v2 (uevIf1 ctx Γ Σ1 Σ2 Σ3 e1 e2 e3 n v2 u n0 u0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e1 e2 e3 : Expr) (n : Number) 
          (v3 : UValue) (u : UEval ctx Γ Σ1 e1 Σ2 (UVNum n)),
        P2 ctx Γ Σ1 e1 Σ2 (UVNum n) u ->
        forall (e : n = Num 0) (u0 : UEval ctx Γ Σ2 e3 Σ3 v3),
        P2 ctx Γ Σ2 e3 Σ3 v3 u0 -> P2 ctx Γ Σ1 (EIf e1 e2 e3) Σ3 v3 (uevIf2 ctx Γ Σ1 Σ2 Σ3 e1 e2 e3 n v3 u e u0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e1 e2 : Expr) (v1 v2 : UValue)
          (u : UEval ctx Γ Σ1 e1 Σ2 v1),
        P2 ctx Γ Σ1 e1 Σ2 v1 u ->
        forall u0 : UEval ctx Γ Σ2 e2 Σ3 v2,
        P2 ctx Γ Σ2 e2 Σ3 v2 u0 -> P2 ctx Γ Σ1 (ESeq e1 e2) Σ3 v2 (uevESeq ctx Γ Σ1 Σ2 Σ3 e1 e2 v1 v2 u u0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e : Expr) (v : UValue) (UR : URefRead) 
          (UW : URefWrite) (u : UEval ctx Γ Σ1 e Σ2 v),
        P2 ctx Γ Σ1 e Σ2 v u ->
        forall m : MkURef Σ2 v (Σ3, UR, UW),
        P2 ctx Γ Σ1 (ERef e) Σ3 (UVRef UR UW) (uevRef ctx Γ Σ1 Σ2 Σ3 e v UR UW u m)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e : Expr) (UR : URefRead) (UW : URefWrite) 
          (v : UValue) (u : UEval ctx Γ Σ1 e Σ2 (UVRef UR UW)),
        P2 ctx Γ Σ1 e Σ2 (UVRef UR UW) u ->
        forall u0 : URefReadEval UR ctx Γ Σ2 Σ3 v,
        P2 ctx Γ Σ1 (EDeref e) Σ3 v (uevDeref ctx Γ Σ1 Σ2 Σ3 e UR UW v u u0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 Σ4 : Heap) (e1 e2 : Expr) (UR : URefRead) 
          (UW : URefWrite) (v : UValue) (u : UEval ctx Γ Σ1 e1 Σ2 (UVRef UR UW)),
        P2 ctx Γ Σ1 e1 Σ2 (UVRef UR UW) u ->
        forall u0 : UEval ctx Γ Σ2 e2 Σ3 v,
        P2 ctx Γ Σ2 e2 Σ3 v u0 ->
        forall u1 : URefWriteEval UW ctx Γ Σ3 v Σ4,
        P2 ctx Γ Σ1 (EAsn e1 e2) Σ4 v (uevAsn ctx Γ Σ1 Σ2 Σ3 Σ4 e1 e2 UR UW v u u0 u1)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap) (x : String) (e : Expr) (F : UFun)
          (e0 : F = MkUFun (ustack Γ) x e), P2 ctx Γ Σ (EFun x e) Σ (UVFun F) (uevFun ctx Γ Σ x e F e0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 Σ4 : Heap) (ef e : Expr) (F : UFun) 
          (lv1 lv2 : UValue) (u : UEval ctx Γ Σ1 ef Σ2 (UVFun F)),
        P2 ctx Γ Σ1 ef Σ2 (UVFun F) u ->
        forall u0 : UEval ctx Γ Σ2 e Σ3 lv1,
        P2 ctx Γ Σ2 e Σ3 lv1 u0 ->
        forall u1 : UFunEval F ctx Γ Σ3 lv1 Σ4 lv2,
        P0 F ctx Γ Σ3 lv1 Σ4 lv2 u1 ->
        P2 ctx Γ Σ1 (EApp ef e) Σ4 lv2 (uevApp ctx Γ Σ1 Σ2 Σ3 Σ4 ef e F lv1 lv2 u u0 u1)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ : Heap), P2 ctx Γ Σ ENil Σ UVNil (uevNil ctx Γ Σ)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (e1 e2 : Expr) (v1 v2 : UValue)
          (u : UEval ctx Γ Σ1 e1 Σ2 v1),
        P2 ctx Γ Σ1 e1 Σ2 v1 u ->
        forall u0 : UEval ctx Γ Σ2 e2 Σ3 v2,
        P2 ctx Γ Σ2 e2 Σ3 v2 u0 ->
        P2 ctx Γ Σ1 (ECons e1 e2) Σ3 (UVCons (MkUHead v1) (MkUTail v2)) (uevCons ctx Γ Σ1 Σ2 Σ3 e1 e2 v1 v2 u u0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (UHd : UHead) (UTl : UTail) 
          (e : Expr) (uv : UValue) (u : UEval ctx Γ Σ1 e Σ2 (UVCons UHd UTl)),
        P2 ctx Γ Σ1 e Σ2 (UVCons UHd UTl) u ->
        forall u0 : UHeadEval UHd ctx Γ Σ2 Σ3 uv,
        P2 ctx Γ Σ1 (EHead e) Σ3 uv (uevHead ctx Γ Σ1 Σ2 Σ3 UHd UTl e uv u u0)) ->
       (forall (ctx : Context) (Γ : Stack) (Σ1 Σ2 Σ3 : Heap) (UHd : UHead) (UTl : UTail) 
          (e : Expr) (uv : UValue) (u : UEval ctx Γ Σ1 e Σ2 (UVCons UHd UTl)),
        P2 ctx Γ Σ1 e Σ2 (UVCons UHd UTl) u ->
        forall u0 : UTailEval UTl ctx Γ Σ2 Σ3 uv,
        P2 ctx Γ Σ1 (ETail e) Σ3 uv (uevTail ctx Γ Σ1 Σ2 Σ3 UHd UTl e uv u u0)) ->
       (forall (l : LFun) (c : Context) (s : Stack) (h : Heap) (l0 : LValue) (h0 : Heap) (l1 : LValue)
         (l2 : LFunEval l c s h l0 h0 l1), P l c s h l0 h0 l1 l2) /\


(forall (u : UFun) (c : Context) (s : Stack) (h : Heap) (u0 : UValue) (h0 : Heap) (u1 : UValue)
         (u2 : UFunEval u c s h u0 h0 u1), P0 u c s h u0 h0 u1 u2) /\
(forall (c : Context) (s : Stack) (h : Heap) (e : Expr) (h0 : Heap) (l : LValue) (l0 : LEval c s h e h0 l),
       P1 c s h e h0 l l0) /\
(forall (c : Context) (s : Stack) (h : Heap) (e : Expr) (h0 : Heap) (u : UValue) (u0 : UEval c s h e h0 u),
       P2 c s h e h0 u u0).
Proof.
intros.
split.
apply LFunEval_mutind with (P:=P)(P0:=P0)(P1:=P1)(P2:=P2); assumption.
split.
apply UFunEval_mutind with (P:=P)(P0:=P0)(P1:=P1)(P2:=P2); assumption.
split.
apply LEval_mutind with (P:=P)(P0:=P0)(P1:=P1)(P2:=P2); assumption.
apply UEval_mutind with (P:=P)(P0:=P0)(P1:=P1)(P2:=P2); assumption.
Qed.

(* ******************************************************************************** *)

Lemma UpgradePublic : forall v, Upgrade L v = v.
Proof.
destruct v; unfold Upgrade; simpl; rewrite Lub_Public; trivial.
Qed.

Lemma UpgradeSecret : forall v, LabelOf (Upgrade H v) = H.
Proof.
unfold Upgrade; unfold LabelOf; destruct v; case l; trivial.
Qed.

(* ******************************************************************************** *)

Lemma MkHeap_Equal : 
  forall Σ, MkHeap (lheap Σ) (uheap Σ) (mheap Σ) = Σ.
Proof.
destruct Σ.
auto.
Qed.

Lemma MkStack_Equal : 
  forall Γ, MkStack (lstack Γ) (ustack Γ) (mstack Γ) = Γ.
Proof.
destruct Γ.
auto.
Qed.

Lemma LabelOf_Upgrade_Secret : 
  forall lv, LabelOf (Upgrade H lv) = H. 
Proof.
intro lv.
  destruct lv; simpl; apply Lub_Secret.
Qed.