Require Import Basics.

(* Syntax *)

Inductive Expr : Set :=
  | ENum : Number -> Expr
  | EVar : String -> Expr

  | ERef : Expr -> Expr
  | EDeref : Expr -> Expr
  | EAsn : Expr -> Expr -> Expr

  | EIf : Expr -> Expr -> Expr -> Expr
  | ESeq : Expr -> Expr -> Expr

  (* Only one variable for proofs *)
  | EFun : String -> Expr -> Expr
  | EApp : Expr -> Expr -> Expr

  | ENil : Expr
  | ECons : Expr -> Expr -> Expr
  | EHead : Expr -> Expr 
  | ETail : Expr -> Expr

  | ELib : String -> Expr.

(* Label ******************************************* *)

Inductive Label : Set := L | H.

(* Model language  ********************************* *)

(* Label terms, κ *)
Inductive LTerm : Set :=
  | LVar : String -> LTerm
  | LLub : LTerm -> LTerm -> LTerm
  | LLabel : Label -> LTerm.

(* Security contexts *)
Definition Context := Label.
Definition ContextModel := LTerm.

(* Label variables, α *)
Definition Alpha := String.

(* Abstract names, π *)
Definition Pi := String.

(* Unlabel Models *)
Inductive UModel : Set :=
  | UMPrim : Alpha -> UModel
  | UMRef  : Alpha -> Pi -> UModel
  | UMFun  : Alpha -> Pi -> UModel
  | UMList : Alpha -> UModel -> UModel
                         
(* Label Models *)
with LModel : Set :=
  | LMPrim : LTerm -> LModel 
  | LMRef  : LTerm -> LModel -> UModel -> LModel
  | LMFun  : LTerm -> UModel -> LModel -> list EffC -> LModel
  | LMList : LTerm -> LModel -> LModel
                                                     
(* Effect Constraints *)
with EffC : Set :=
  | ECRefRead  : Pi -> UModel -> EffC
  | ECRefWrite : Pi -> ContextModel -> LModel -> EffC
  | ECApp      : Pi -> ContextModel -> LModel -> UModel -> EffC
  | ECAsn      : ContextModel -> Alpha -> LTerm -> EffC.

(* ******************************************************************************** *)

Definition lub (l1 : Label) (l2 : Label) : Label :=
  match l1 with
  | L => l2
  | H => H
  end.

Definition LTermOf (lm : LModel) : LTerm :=
  match lm with
  | LMPrim κ => κ
  | LMRef κ _ _ => κ
  | LMFun κ _ _ _ => κ
  | LMList κ  _ => κ
  end.

Definition LVarOf (um : UModel) : Alpha :=
  match um with 
  | UMPrim α => α
  | UMRef α _ => α
  | UMFun α _ => α
  | UMList α _ => α
  end.

(* ******************************************************************************** *)

Lemma Lub_Symm : forall l1 l2, lub l1 l2 = lub l2 l1.
Proof.
destruct l1; destruct l2; simpl; trivial.
Qed.

Lemma Lub_Secret : forall l, lub l H = H.
Proof.
destruct l; simpl; trivial.
Qed.

Lemma Lub_Public : forall l, lub l L = l.
Proof.
destruct l; simpl; trivial.
Qed.

Inductive CanFlow : Label -> Label -> Prop :=
  | cfLow : forall l, CanFlow L l
  | cfHigh : CanFlow H H.




